/* 
 * eclipse-winscp-upload-script
 * Copyright (C) 2012  Stefan Kleeschulte
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Additional information about this script is available under the following
 * URL:
 * http://rays-blog.de/2012/05/05/94/use-winscp-to-upload-files-using-eclipses-autobuild-feature/
 */


/*******************************************************************************
*  C O N F I G U R A T I O N                                                   *
*******************************************************************************/

/*
 * -----------------------------------------------------------------------------
 * dir_assignment
 * -----------------------------------------------------------------------------
 * 
 * Assignment of local directories to remote directories. Local directories are
 * relative to the project root. Use (back-)slashes as shown in the examples.
 * 
 * If for a folder there is one assignment, and for one of its parent folders
 * there is another assignment, it will always be used the most specific
 * assignment that fits to a path.
 * 
 * Examples:
 * 
 * var dir_assignment = [
 *   ["\\files","/files/project"],
 *   ["\\htdocs","/htdocs/project"]
 * ];
 * 
 * var dir_assignment = [
 *   ["\\subfolder1\\subfolder2","/remote/path"]
 * ];
 * 
 * var dir_assignment = [
 *   ["\\subfolder1","/remote/path1"],
 *   ["\\subfolder1\\subfolder2","/remote/path2"]
 * ];
 */

var dir_assignment = [
  ["\\files","/files/project"],
  ["\\htdocs","/htdocs/project"]
];

/*
 * -----------------------------------------------------------------------------
 * winscp_com
 * -----------------------------------------------------------------------------
 * 
 * Path to WinSCP.com, relative to the "Base Directory" specified in the
 * builders launch configuration
 * 
 * Examples:
 * 
 * var winscp_com = 'C:\\Program Files (x86)\\WinSCP\\WinSCP.com';
 * var winscp_com = 'WinSCP.com';
 */

var winscp_com = 'WinSCP.com';

/*
 * -----------------------------------------------------------------------------
 * winscp_open_arguments
 * -----------------------------------------------------------------------------
 * 
 * Arguments for WinSCP's open command. These specify the connection settings.
 * Supported Protocols are FTP, FTPS (FTP-SSL), SFTP and SCP.
 * 
 * See WinSCP's scripting documentation for an overview of available arguments:
 * http://winscp.net/eng/docs/scriptcommand_open
 * 
 * A password needs to be specified in cleartext here. However, as long as the
 * connection is specified in the format [protocol]://[user]:[password]@[server]
 * the password will be starred out in the console output.
 * 
 * Examples:
 * 
 * FTP: 
 * var winscp_open_arguments = 'ftp://user:password@server.tld';
 * 
 * FTPS:
 * var winscp_open_arguments = 'ftps://user:password@server.tld -explicitssl -certificate="00:11:22:33:44:55:66:77:88:99:aa:bb:cc:dd:ee:ff:a0:b1:c2:d3"';
 * (To find out the certificate's fingerprint, connect to the server using
 * WinSCP GUI (WinSCP.exe) and go to Commands -> Server/Protocol Information.)
 */

var winscp_open_arguments = 'ftp://user:password@server.tld';

/*
 * -----------------------------------------------------------------------------
 * synchronize_on_autobuild_failure
 * -----------------------------------------------------------------------------
 * 
 * Boolean variable, determines whether on an autobuild failure a full sync
 * shall be performed. Explanation: When build.js is triggered by eclipses
 * autobuild feature, only added, changed and removed files are passed to
 * WinSCP by the script. If for some reason WinSCP fails to execute the given
 * commands (meaning WinSCP's exit code is not 0), the script can
 * automatically perform a full sync of the specified directories using
 * WinSCP's synchronize feature.
 * 
 * Examples:
 * 
 * var synchronize_on_autobuild_failure = true; // default
 * var synchronize_on_autobuild_failure = false;
 * 
 */

var synchronize_on_autobuild_failure = true;

/*
 * -----------------------------------------------------------------------------
 * winscp_console_mode
 * -----------------------------------------------------------------------------
 * 
 * Normally this script uses StdIn and StdOut to control WinSCP. This is faster
 * as no temporary WinSCP batch file has to be created. However, if you
 * encounter problems you might want to try setting this to true.
 * 
 * Examples:
 * 
 * var winscp_console_mode = false; // default
 * var winscp_console_mode = true;
 */

var winscp_console_mode = false;

/*
 * -----------------------------------------------------------------------------
 * explicitly_delete_folder_contents
 * -----------------------------------------------------------------------------
 * 
 * Normally, when a non empty folder is removed, only one command for deletion
 * of this folder is passed to WinSCP. However, if this for some reason does not
 * work, you can have this script pass explicit commands to WinSCP to remove all
 * subfolders and files first.
 * 
 * Examples:
 * 
 * var explicitly_delete_folder_contents = false; // default
 * var explicitly_delete_folder_contents = true;
 */
var explicitly_delete_folder_contents = false;



/*******************************************************************************
*  S C R I P T                                                                 *
*******************************************************************************/

// read action and project_dir from arguments
if (WScript.Arguments.length < 2) {
	// quit with error
	WScript.Echo('Wrong arguments - no action/project dir specified.');
	WScript.Quit(1);
} else {
	var action = WScript.Arguments.Item(0);
	var project_dir = WScript.Arguments.Item(1);
}

// determine what to do:
// sync- or autobuild-mode, depending on action
if (action == '-autobuild') {
	
	var autobuild_result = autobuild();
	if (autobuild_result != 0 && synchronize_on_autobuild_failure) {
		WScript.Echo('autobuild failed (' + autobuild_result + '), starting full sync...');
		WScript.Quit(sync());
	} else {
		WScript.Quit(autobuild_result);
	}
	
} else if (action == '-sync') {
	
	WScript.Quit(sync());
	
} else {
	
	// unknown action -> quit with error
	WScript.Echo('Wrong arguments - unknown action "' + action + '".');
	WScript.Quit(1);
	
}

// replicate local changes on server
function autobuild() {
	
	var paths = new Object();
	paths.added			= [];
	paths.changed		= [];
	paths.removed		= [];
	paths.files			= [];
	paths.dirs			= [];
	paths.added_files	= [];
	paths.added_dirs	= [];
	paths.changed_files	= [];
	paths.changed_dirs	= [];
	paths.removed_files	= [];
	paths.removed_dirs	= [];
	
	if (WScript.Arguments.length < 4) {
		// quit with error
		WScript.Echo('Wrong arguments - no paths or sourcefile specified.');
		WScript.Quit(1);
	} else if (WScript.Arguments.Item(2) == '-s') {
		// read paths from sourcefile
		
		var sourcefile = WScript.Arguments.Item(3);
		// read sourcefile contents
		var fso = new ActiveXObject('Scripting.FilesystemObject');
		if (fso.FileExists(sourcefile)) {
			var tf = fso.OpenTextFile(sourcefile, 1, false);
			while (!tf.AtEndOfStream) {
				var line = tf.ReadLine();
				var type = line.substr(0, 1);
				var data = line.substring(3, line.length - 1);
				data = data.split('" "');
				switch (type) {
					case 'a':
						paths.added = paths.added.concat(data);
						break;
					case 'c':
						paths.changed = paths.changed.concat(data);
						break;
					case 'r':
						paths.removed = paths.removed.concat(data);
						break;
					case 'f':
						paths.files = paths.files.concat(data);
						break;
					case 'd':
						paths.dirs = paths.dirs.concat(data);
				}
			}
			tf.Close();
		} else {
			WScript.Echo('Sourcefile "' + sourcefile + '" does not exist.');
			WScript.Quit(1);
		}
		
	} else {
		// read paths from arguments
		
		var current_list = '';
		for (var i = 2; i < WScript.Arguments.length; i++) {
			switch (WScript.Arguments.Item(i)) {
				case '-a':
					current_list = 'added';
					break;
				case '-c':
					current_list = 'changed';
					break;
				case '-r':
					current_list = 'removed';
					break;
				case '-f':
					current_list = 'files';
					break;
				case '-d':
					current_list = 'dirs';
					break;
				default:
					if (current_list != '') {
						paths[current_list].push(WScript.Arguments.Item(i));
					} else {
						WScript.Echo('Wrong arguments - path arguments do not match required format.');
						WScript.Quit(1);
					}
			}
		}
		
	}
	
	//sort dir assignments by string length of local path in descending order
	dir_assignment.sort(function(a, b) { return b[0].length - a[0].length; });
	
	//split path arrays into separate arrays for files and dirs
	var types = ['added', 'changed', 'removed'];
	for (var i = 0; i < types.length; i++) {
		var type = types[i];
		for (var j = 0; j < paths[type].length; j++) {
			if (indexOf(paths[type][j], paths['files']) > -1) {
				paths[type+'_files'].push(paths[type][j]);
			} else {
				paths[type+'_dirs'].push(paths[type][j]);
			}
		}
	}
	
	// create list of winscp commands
	var commands = [
		'option batch abort',
		'option confirm off',
		'open ' + winscp_open_arguments,
		'option transfer binary'
	];
	
	// check if a directory or file was moved or renamed
	var moves = movedOrRenamed(paths);
	if (moves) {
		// move moved files and dirs
		for (var i = 0; i < moves.length; i++) {
			commands.push('mv "' + remotePath(moves[i][0]) + '" "' + remotePath(moves[i][1]) + '"');
		}
		
		// upload changed files
		for (var i = 0; i < paths.changed_files.length; i++) {
			commands.push('put "' + paths.changed_files[i] + '" "' + remotePath(paths.changed_files[i]) + '"');
		}
	} else {
		// create added dirs
		paths.added_dirs.sort(function (a, b) { return a.length - b.length; });
		for (var i = 0; i < paths.added_dirs.length; i++) {
			var new_dir = remotePath(paths.added_dirs[i]);
			if (new_dir != false)
				commands.push('mkdir "' + new_dir + '"');
		}
	
		// upload added files
		for (var i = 0; i < paths.added_files.length; i++) {
			var new_file = remotePath(paths.added_files[i]);
			if (new_file != false)
				commands.push('put "' + paths.added_files[i] + '" "' + new_file + '"');
		}
	
		// upload changed files
		for (var i = 0; i < paths.changed_files.length; i++) {
			var file = remotePath(paths.changed_files[i]);
			if (file != false)
				commands.push('put "' + paths.changed_files[i] + '" "' + file + '"');
		}
	
		if (explicitly_delete_folder_contents) {
			// delete removed files
			for (var i = 0; i < paths.removed_files.length; i++) {
				var file = remotePath(paths.removed_files[i]);
				if (file != false)
					commands.push('rm "' + file + '"');
			}
		
			// delete removed dirs
			paths.removed_dirs.sort(function (a, b) { return b.length - a.length; });
			for (var i = 0; i < paths.removed_dirs.length; i++) {
				var dir = remotePath(paths.removed_dirs[i]);
				if (dir != false)
					commands.push('rmdir "' + dir + '"');
			}
		} else {
			var deletions = requiredDeletions(paths);
			
			// delete removed files
			for (var i = 0; i < deletions[0].length; i++) {
				var path = remotePath(deletions[0][i]);
				if (path != false)
					commands.push('rm "' + path + '"');
			}
			
			// delete removed dirs
			for (var i = 0; i < deletions[1].length; i++) {
				var path = remotePath(deletions[1][i]);
				if (path != false)
					commands.push('rmdir "' + path + '"');
			}
		}
	}
	
	// winscp exit command
	commands.push('exit');
	
	// commit changes to server
	return execWinSCPCommands(commands);
	
}

// sync server with local dirs
function sync() {
	
	// sort dir assignments by string length of local path in ascending order
	dir_assignment.sort(function(a, b) { return a[0].length - b[0].length; });
	
	// create list of winscp commands
	var commands = [
		'option batch abort',
		'option confirm off',
		'open ' + winscp_open_arguments,
		'option transfer binary'
	];
	
	// add a synchronize command for each directory assignment
	for (var i = 0; i < dir_assignment.length; i++) {
		commands.push('synchronize remote -delete -criteria=either "' + project_dir + dir_assignment[i][0] + '" "' + dir_assignment[i][1] + '"');
	}
	
	// winscp exit command
	commands.push('exit');
	
	// sync server
	return execWinSCPCommands(commands);
	
}


// Returns an array of pairs of old and new location of files and directories
// which have been moved (or renamed), or false if nothing has been moved (or
// renamed).
// This function assumes that eclipse only allows to move files and directories
// from the same source directory to the same destination directory in one
// operation (meaning before autobuild is triggered), which is true for 3.7.1
// (Indigo).
function movedOrRenamed(paths) {
	if (paths.added.length != 0 && paths.added.length == paths.removed.length) {
		// sort added and removed array by string length in ascending order
		paths.added.sort(function(a, b) { return a.length - b.length; });
		paths.removed.sort(function(a, b) { return a.length - b.length; });
		// check if all paths from added and removed array differ only by the
		// first ("base") part of the path
		var added_base = paths.added[0].split('\\');
		var added_base_elements = added_base.length;
		added_base = added_base.slice(0, added_base_elements - 1).join('\\');
		var added_base_length = added_base.length;
		var removed_base = paths.removed[0].split('\\');
		var removed_base_elements = removed_base.length;
		removed_base = removed_base.slice(0, removed_base_elements - 1).join('\\');
		var removed_base_length = removed_base.length;
		for (var i = 1; i < paths.added.length; i++) {
			if (paths.added[i].substring(added_base_length) != paths.removed[i]
					.substring(removed_base_length)) {
				return false;
			}
		}
		// check if added_base and removed_base are both in server synced dirs
		// (according to dir_assignment)
		if (remotePath(added_base) == false || remotePath(removed_base) == false) {
			return false;
		}
		// return files and dirs to be moved
		var moves = [];
		for (var i = 0; i < paths.added.length; i++) {
			var move = [
				paths.removed[i].split('\\').slice(0, removed_base_elements).join('\\'),
				paths.added[i].split('\\').slice(0, added_base_elements).join('\\')
			];
			// check if this 'move' is already in moves array
			var already_there = false;
			for (var j = 0; j < moves.length; j++) {
				if (moves[j][0] == move[0]) {
					already_there = true;
					break;
				}
			}
			if (!already_there) {
				moves.push(move);
			}
		}
		return moves;
	}
	return false;
}

// Returns a list of files and directories which have to be deleted if
// the contents of directories do not have to be deleted explicitly.
function requiredDeletions(paths) {
	// sort removed_dirs array by string length in ascending order
	paths.removed_dirs.sort(function (a, b) { return a.length - b.length; });
	var deletions = [[], []];
	// for each file, check if it is in a directory which is removed too
	for (var i = 0; i < paths.removed_files.length; i++) {
		var in_removed_dir = false;
		for (var j = 0; j < paths.removed_dirs.length; j++) {
			if (paths.removed_files[i].substr(0, paths.removed_dirs[j].length + 1) == paths.removed_dirs[j] + '\\') {
				in_removed_dir = true;
				break;
			}
		}
		if (!in_removed_dir) {
			deletions[0].push(paths.removed_files[i]);
		}
	}
	// for each directory, check if it is in another directory which is
	// removed too
	for (var i = paths.removed_dirs.length - 1; i >= 0; i--) {
		var in_removed_dir = false;
		for (var j = 0; j < i; j++) {
			if (paths.removed_dirs[i].substr(0, paths.removed_dirs[j].length + 1) == paths.removed_dirs[j] + '\\') {
				in_removed_dir = true;
				break;
			}
		}
		if (!in_removed_dir) {
			deletions[1].push(paths.removed_dirs[i]);
		}
	}
	return deletions;
}

// Returns the position of an element in an array, or -1 if the array
// does not contain the element. (Does not work for elements of object
// type - including arrays.)
function indexOf(element, array) {
	for (var i = 0; i < array.length; i++) {
		if (array[i] === element)
			return i;
	}
	return -1;
}

// Returns the path to a file or folder on the ftp server according to
// given dir assignments.
function remotePath(path) {
	// remove project_dir part of path
	path = path.substring(project_dir.length);
	// search for fitting dir assignment
	var ldir = '';
	for (var i = 0; i < dir_assignment.length; i++)  {
		ldir = path.substr(0, dir_assignment[i][0].length);
		if (ldir == dir_assignment[i][0]) {
			// return remote path
			return dir_assignment[i][1] + path.substring(ldir.length).replace(/\\/g, '/');
		}
	}
	return false;
}

// Generates a name for the WinSCP script file
function getWinSCPScriptName() {
	var date = new Date();
	var month = date.getMonth() + 1;
	var day = date.getDate();
	var hours = date.getHours();
	var minutes = date.getMinutes();
	var seconds = date.getSeconds();
	if (month < 10) month = '0' + month;
	if (day < 10) day = '0' + day;
	if (hours < 10) hours = '0' + hours;
	if (minutes < 10) minutes = '0' + minutes;
	if (seconds < 10) seconds = '0' + seconds;
	
	return 'autobuild_' + month + '-' + day + '-' + date.getFullYear() + '_' + hours + '-' + minutes + '-' + seconds + '.txt';
}

// execute WinSCP commands
// returns 0 on success, otherwise a value greater than 0
function execWinSCPCommands(commands) {
	var error = false;
	if (!winscp_console_mode) {
		// pass commands to WinSCP using script file
		var filename = getWinSCPScriptName();
		var fso = new ActiveXObject('Scripting.FilesystemObject');
		var tf = fso.CreateTextFile(filename, true);
		for (var i = 0; i < commands.length; i++) {
			tf.WriteLine(commands[i]);
		}
		tf.Close();
		var shell = WScript.CreateObject('WScript.Shell');
		WScript.Echo('Starting WinSCP...');
		var shellExec = shell.Exec('"' + winscp_com + '" /script=' + filename);
		while (!shellExec.StdOut.AtEndOfStream) {
			var line = shellExec.StdOut.ReadLine();
			WScript.Echo(line);
			if (line.substring(line.length - 8) == ' failed.') {
				error = true;
			}
			//WScript.Sleep(10);
		}
		if (fso.FileExists(filename)) {
			fso.DeleteFile(filename);
		}
	} else {
		// pass commands to WinSCP using 'interactive' console
		var shell = WScript.CreateObject('WScript.Shell');
		WScript.Echo('Starting WinSCP...');
		var shellExec = shell.Exec(winscp_com);
		var output = '';
		var lastchar = '';
		var escape_pwd = 0;
		var hide_output = false;
		var i = 0;
		while (!shellExec.StdOut.AtEndOfStream) {
			lastchar = shellExec.StdOut.Read(1);
			if (!hide_output) {
				WScript.StdOut.Write(lastchar);
			}
			if (lastchar == '\n') {
				// Windows 8 might use \n for newlines (instead of \r\n) - see http://en.wikipedia.org/wiki/Newline
				if (output.substring(output.length - 9) == ' failed.\r' || output.substring(output.length - 8) == ' failed.') {
					error = true;
				}
				output = '';
			} else {
				output += lastchar;
				if (escape_pwd == 1 && output.substring(output.length - 3) == '://' && lastchar != '\n') {
					output = '';
					escape_pwd = 2;
				} else if(escape_pwd == 2 && output.substring(output.length - 1) == ':' && lastchar != '\n') {
					hide_output = true;
					WScript.StdOut.Write('***');
					output = '';
					escape_pwd = 3;
				} else if(escape_pwd == 3 && output.substring(output.length - 1) == '@' && lastchar != '\n') {
					hide_output = false;
					WScript.StdOut.Write('@');
					escape_pwd = 0;
				} else if (output == 'winscp> ') {
					if (commands[i].substr(0, 5) == 'open ') {
						escape_pwd = 1;
					} else {
						escape_pwd = 0;
					}
					shellExec.StdIn.Write(commands[i] + '\r\n');
					i++;
				}
			}
		}
	}
	if (shellExec.ExitCode == 0 && !error) {
		WScript.Echo('WinSCP finished without errors.');
	} else if (shellExec.ExitCode == 0 && error) {
		WScript.Echo('WinSCP finished with errors.');
	}
	else {
		WScript.Echo('WinSCP finished with errors - exit code ' + shellExec.ExitCode + '.');
	}
	return (shellExec.ExitCode == 0 && error) ? 1 : shellExec.ExitCode;
}